var $ = jQuery.noConflict()

var draftback = {}

draftback.indexedDB = {}
draftback.indexedDB.db = null

draftback.pendingRevisions = {}

draftback.getParameterByName = function(url, name) {
  var match = RegExp('[?&]' + name + '=([^&]*)').exec(url);
  return match && decodeURIComponent(match[1].replace(/\+/g, ' '));
}

draftback.addPendingRevision = function(data, requestId) {
  draftback.pendingRevisions[requestId] = data
}

draftback.getPendingRevision = function(requestId) {
  return draftback.pendingRevisions[requestId]
}

draftback.deletePendingRevision = function(requestId) {
  delete draftback.pendingRevisions[requestId]
}

Array.prototype.insert = function(element, index) {
  this.splice(index, 0, element)
}
    
Array.prototype.delete = function(startIndex, endIndex) {
  return this.splice(startIndex, (endIndex - startIndex) + 1)
}

Array.prototype.last = function() {
  return this[this.length - 1];
}

draftback.kills = {}

draftback.shouldKill = function(docId) {
  if (draftback.kills[docId]) {
    delete draftback.kills[docId]
    return true
  } else {
    return false
  }
}

draftback.authorMaps = {}

draftback.autoincrements = {}
draftback.windows = {}

draftback.writing_sessions = {}
draftback.timelines = {}

draftback.render = function(characters, entry, builder, cllbck) {
  var character_limit = 1000 // Magic number change in two places
  var slop = 500
  var should_change_viewing_window_to = false
  var old_window = draftback.windows[builder.docId]
  var current_window_start_i = old_window[0]
  var current_window_end_i = old_window[1]
  
  char_length = characters.length
  
  if (entry['ty'] == "is") {
    var start_i = entry['ibi'] - 1
    var end_i = start_i + entry['s'].length
  } else if (entry['ty'] == "ds") {
    var start_i = entry['si'] - 1
    var end_i = entry['ei']
  }
  
  if (current_window_end_i - current_window_start_i > character_limit + slop) {
    current_window_start_i += slop / 2
    current_window_end_i -= slop / 2
  }
  
  if (end_i - start_i > character_limit) {
    var left_cut = current_window_start_i
    var right_cut = current_window_end_i
  } else {
    if (start_i < current_window_start_i) {
      var new_start = Math.max(start_i - 100, 0)
      draftback.windows[builder.docId] = [new_start, new_start + character_limit]
    } else if (end_i > current_window_end_i) {
      var new_end = Math.min(end_i + 100, char_length - 1)
      draftback.windows[builder.docId] = [new_end - character_limit, new_end]
    } else {
      if (entry['ty'] == "is") {
        draftback.windows[builder.docId] = [current_window_start_i, current_window_end_i + entry['s'].length]
      } else if (entry['ty'] == "ds") {
        should_change_viewing_window_to = [current_window_start_i, current_window_end_i - (end_i - start_i)]
      }
    }
  }
  var builder_viewing_window = draftback.windows[builder.docId]
  var left_cut = builder_viewing_window[0]
  var right_cut = builder_viewing_window[1]
  
  var html = ""
  
  if (char_length < 35000) {
    html = "<div class='progress'><p>"
    $.each(characters, function(i, char) {
      if (!char) return true
      
      if (i == left_cut) {
        html += "<div class='window'><p>"
      }
      if (i == start_i) {
        html += (entry['ty'] == "is" ? "<ins>" : "<del>")
      }
      if (char.s == "\n") {
        html += "</p><p>"
      } else {
        html += char.s
      }
      if (i == end_i) {
        html += (entry['ty'] == "is" ? "</ins>" : "</del>")
      }
      if (i == right_cut || i == char_length - 1) {
        html += "</p></div>"
      }
    })
    html += "</p></div>"
    html = html.replace(/<p><\/p>/g, "<p>&nbsp;</p>")
  }
  
  html += "<div class='content'>"
  
  var delta_start_point = start_i - left_cut
  var delta_end_point = end_i - left_cut
  
  $.each(characters.slice(left_cut, right_cut + 1), function(i, char) {
    if (!char) return true
    
    if (i == delta_start_point) {
      html += (entry['ty'] == "is" ? "<ins>" : "<del>")
    }
    
    if (char.s == "\n") {
      html += "<br>"
    } else {
      html += char.s
    }
    
    if (i == delta_end_point) {
      html += (entry['ty'] == "is" ? "</ins>" : "</del>")
    }
  })
  
  if (should_change_viewing_window_to) {
    draftback.windows[builder.docId] = should_change_viewing_window_to
  }
  
  if (!draftback.writing_sessions[builder.docId]) {
    draftback.writing_sessions[builder.docId] = {}
  }
  
  if (!draftback.timelines[builder.docId]) {
    draftback.timelines[builder.docId] = []
  }
  
  var tmln = {
    timestamp: entry.timestamp,
    character_count: char_length,
    start_i: start_i
  }
  
  var last_tmln = draftback.timelines[builder.docId].last()
  if (last_tmln) {
    var timelines_are_different = false
    for (var tmln_key in last_tmln) {
      if (last_tmln.hasOwnProperty(tmln_key)) {
        if (last_tmln[tmln_key] != tmln[tmln_key]) {
          timelines_are_different = true
        }
      }
    }
    if (timelines_are_different) {
      draftback.timelines[builder.docId].push(tmln)
    }
  } else {
    draftback.timelines[builder.docId].push(tmln)
  }
  
  var current_session = draftback.writing_sessions[builder.docId][entry.uid]
  if (current_session) {
    var continuation_interval = 1000 * 60 * 10 // 10 minutes
    if ((entry.timestamp - current_session.end) < continuation_interval) {
      current_session['end'] = entry.timestamp
      current_session['end_revision'] = entry.revno
      current_session['count'] += 1
    } else {
      var sess = {
        seq: draftback.autoincrements[builder.docId],
        docId: builder.docId,
        uid: entry.uid,
        start_revision: '' + current_session.start_revision,
        end_revision: '' + current_session.end_revision,
        start_timestamp: current_session.start,
        end_timestamp: current_session.end,
        duration: current_session.end - current_session.start,
        revisions_count: current_session.count
      }
      draftback.indexedDB.addSession(sess)
      delete draftback.writing_sessions[builder.docId][entry.uid]
    }
  } else {
    for (var user_id in draftback.writing_sessions[builder.docId]) {
      if (draftback.writing_sessions[builder.docId].hasOwnProperty(user_id)) {
        var current_session = draftback.writing_sessions[builder.docId][user_id]
        var sess = {
          seq: draftback.autoincrements[builder.docId],
          docId: builder.docId,
          uid: user_id,
          start_revision: '' + current_session.start_revision,
          end_revision: '' + current_session.end_revision,
          start_timestamp: current_session.start,
          end_timestamp: current_session.end,
          duration: current_session.end - current_session.start,
          revisions_count: current_session.count
        }
        draftback.indexedDB.addSession(sess)
        delete draftback.writing_sessions[builder.docId][user_id]
      }
    }
    draftback.writing_sessions[builder.docId][entry.uid] = {'start': entry.timestamp, 'start_revision': entry.revno, 'end': entry.timestamp, 'end_revision': entry.revno, 'count': 1}
  }

  var rev = {
    seq: draftback.autoincrements[builder.docId]++,
    docId: builder.docId,
    rendered: html,
    timestamp: entry.timestamp,
    revno: entry.revno
  }
  draftback.indexedDB.addRevision(rev, cllbck)
}

draftback.buildRevisions = function(docId, raw_revisions) {
  chrome.tabs.query({url: '*://docs.google.com/*/' + docId + '/edit'}, function(tabs) {
    chrome.tabs.sendMessage(tabs[0].id, {msg: 'clearing'}, function(response) {});
  });
  
  draftback.clearAllRevisionsByDocId(docId, function() {
    chrome.tabs.query({url: '*://docs.google.com/*/' + docId + '/edit'}, function(tabs) {
      chrome.tabs.sendMessage(tabs[0].id, {msg: 'cleared'}, function(response) {});
    });
    
    var raw = raw_revisions.split("\n").pop()
    var changelog = JSON.parse(raw).changelog
    var totalChangelogSize = changelog.length
    var characters = []
    draftback.autoincrements[docId] = 0
    draftback.windows[docId] = [0, 1000] // Magic number change in two places
  
    var handlers = {
      'is': function insert(data) {
        // console.log('is')
        var insertBeginIndex = data.ibi,
              stringToInsert = data.s
      
        var charactersToInsert = $.map(stringToInsert, function(s, i) { return {"s": s} })
      
        for (var i = 0; i < stringToInsert.length; i++) {
          characters.insert(charactersToInsert[i], (insertBeginIndex - 1) + i)
        }
      },
      'ds': function del(data) {
        // console.log('ds')
        var deleteStartIndex = data.si,
              deleteEndIndex = data.ei
            
        var deleted = characters.delete(deleteStartIndex - 1, deleteEndIndex - 1)
      },
      'mlti': function multi(data, clb) {
        var i = 0
        async.eachSeries(data.mts, function(subcmd, mlticb) {
          if (!subcmd.revno) subcmd.revno = [data.revno, i].join('-')
          subcmd.timestamp = data.timestamp
          subcmd.uid = data.uid
          dispatch(subcmd, mlticb)
          i++
        }, function() { clb() })
      }
    }
  
    var dispatch = function(ent, theCallback) {
      var handler = handlers[ent.ty]
    
      if (handler) {
        if (ent.ty == 'is') {
          handler(ent)
          draftback.render(characters, ent, {docId: docId}, theCallback)
        } else if (ent.ty == 'ds') {
          draftback.render(characters, ent, {docId: docId}, theCallback)
          handler(ent)
        } else if (ent.ty == 'mlti') {
          handler(ent, theCallback)
        }
      } else {
        theCallback()
      }
    }

    var it = 0
    async.eachSeries(changelog, function(ce, theTopCallback) {
      if (draftback.shouldKill(docId)) return
      var entry = $.extend(ce[0], {timestamp: ce[1], uid: ce[2], revno: ce[3]})
      dispatch(entry, theTopCallback)
      chrome.tabs.query({url: '*://docs.google.com/*/' + docId + '/edit', active: true}, function(tabs) {
        chrome.tabs.sendMessage(tabs[0].id, {msg: 'progress', so_far: (it + 1), out_of: totalChangelogSize}, function(response) {});
      });
      
      if (it + 1 == totalChangelogSize) {
        for (var user_id in draftback.writing_sessions[docId]) {
          if (draftback.writing_sessions[docId].hasOwnProperty(user_id)) {
            var current_session = draftback.writing_sessions[docId][user_id]
            var sess = {
              seq: draftback.autoincrements[docId],
              docId: docId,
              uid: user_id,
              start_revision: '' + current_session.start_revision,
              end_revision: '' + current_session.end_revision,
              start_timestamp: current_session.start,
              end_timestamp: current_session.end,
              duration: current_session.end - current_session.start,
              revisions_count: current_session.count
            }
            draftback.indexedDB.addSession(sess)
            delete draftback.writing_sessions[docId][user_id]
          }
        }
        
        var timelineToSave = draftback.timelines[docId]
        if (timelineToSave.length > 2500) {
          var skip = Math.ceil(timelineToSave.length / 2500)
          var newTimelineToSave = []
          for (var i = 0; i < timelineToSave.length; i += skip) {
            newTimelineToSave.push(timelineToSave[i])
          }
          timelineToSave = newTimelineToSave
        }

        draftback.indexedDB.addTimeline({docId: docId, timeline: timelineToSave})
      }
    
      it++
    })
  })
}

draftback.indexedDB.open = function() {
  var version = 7
  var request = indexedDB.open("revisions", version)

  request.onupgradeneeded = function(e) {
    var db = e.target.result

    e.target.transaction.onerror = draftback.indexedDB.onerror
    if (!db.objectStoreNames.contains("revision")) {
      var store = db.createObjectStore("revision", { autoIncrement: true })
    
      store.createIndex("docseq", ["docId", "seq"], { unique: false })
    }
    
    if (!db.objectStoreNames.contains("session")) {
      var store = db.createObjectStore("session", { autoIncrement: true })
      
      store.createIndex("sessdocseq", ["docId", "seq"])
    }
    
    if (db.objectStoreNames.contains("timeline")) {
      db.deleteObjectStore("timeline")
    }
    var timeline_store = db.createObjectStore("timeline", {keyPath: "docId"});
  }

  request.onsuccess = function(e) {
    draftback.indexedDB.db = e.target.result
  }

  request.onerror = draftback.indexedDB.onerror
}

draftback.indexedDB.addRevision = function(data, cb) {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["revision"], "readwrite")
  var store = trans.objectStore("revision")
  
  var request = store.put(data)

  request.onsuccess = function(e) { 
    // console.log('commit to db')
    cb()
  }

  request.onerror = function(e) {
    console.log(e.value)
  }
}

draftback.indexedDB.addSession = function(data, cb) {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["session"], "readwrite")
  var store = trans.objectStore("session")
  
  var request = store.put(data)

  request.onsuccess = function(e) { 
    // console.log('commit to db')
    // cb()
  }

  request.onerror = function(e) {
    console.log(e.value)
  }
}

draftback.indexedDB.addTimeline = function(data, cb) {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["timeline"], "readwrite")
  var store = trans.objectStore("timeline")
  
  var request = store.put(data)

  request.onsuccess = function(e) { 
    // console.log('commit to db')
    // cb()
  }

  request.onerror = function(e) {
    console.log(e.value)
  }
}

draftback.indexedDB.getAllRevisions = function() {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["revision"], "readwrite")
  var store = trans.objectStore("revision")

  var keyRange = IDBKeyRange.lowerBound(0)
  var cursorRequest = store.openCursor(keyRange)

  cursorRequest.onsuccess = function(e) {
    var result = e.target.result
    if(!!result == false)
      return

    result.continue()
  }

  cursorRequest.onerror = draftback.indexedDB.onerror
}

draftback.indexedDB.lastRevisionByDocId = function(docId, callback) {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["revision"], "readonly")
  var store = trans.objectStore("revision")
  var index = store.index("docseq")
  
  var lowerBound = [docId, 0]
  var upperBound = [docId, Number.MAX_VALUE]
  var range = IDBKeyRange.bound(lowerBound, upperBound)
  
  var openCursorRequest = index.openCursor(range, 'prev')
  var maxRevisionObject = null
  
  openCursorRequest.onsuccess = function(event) {
    if (event.target.result) {
      maxRevisionObject = event.target.result.value;
      callback(maxRevisionObject)
    } else {
      callback({docId: docId, seq: 0})
    }
  }

  openCursorRequest.onerror = draftback.indexedDB.onerror  
}

draftback.indexedDB.lastSessionByDocId = function(docId, callback) {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["session"], "readonly")
  var store = trans.objectStore("session")
  var index = store.index("sessdocseq")
  
  var lowerBound = [docId, 0]
  var upperBound = [docId, Number.MAX_VALUE]
  var range = IDBKeyRange.bound(lowerBound, upperBound)
  
  var openCursorRequest = index.openCursor(range, 'prev')
  var maxSessionObject = null
  
  openCursorRequest.onsuccess = function(event) {
    if (event.target.result) {
      maxSessionObject = event.target.result.value;
      callback(maxSessionObject)
    } else {
      callback({docId: docId, seq: 0})
    }
  }

  openCursorRequest.onerror = draftback.indexedDB.onerror  
}

draftback.indexedDB.getRevisionsByDocIdWithOffsetAndLimit = function(docId, offset, limit, cb) {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["revision"], "readwrite") // FIXME: this should be readonly? as should some others?
  var store = trans.objectStore("revision")
  var index = store.index("docseq")
  
  var lowerBound = [docId, offset]
  var upperBound = [docId, offset + limit]
  var range = IDBKeyRange.bound(lowerBound, upperBound)
  
  var array = []
  var cursorRequest = index.openCursor(range, 'next')

  cursorRequest.onsuccess = function(event) {
    var cursor = event.target.result
    if (cursor) {
      array.push(cursor.value)
      cursor.continue()
    } else if (cb) {
      cb(array)
    }
  }

  cursorRequest.onerror = draftback.indexedDB.onerror
  return array
}

draftback.indexedDB.getAllSessionsByDocId = function(docId, cb) {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["session"], "readwrite") // FIXME: this should be readonly? as should some others?
  var store = trans.objectStore("session")
  var index = store.index("sessdocseq")
  
  var lowerBound = [docId, 0]
  var upperBound = [docId, Number.MAX_VALUE]
  var range = IDBKeyRange.bound(lowerBound, upperBound)
  
  var array = []
  var cursorRequest = index.openCursor(range, 'next')

  cursorRequest.onsuccess = function(event) {
    var cursor = event.target.result
    if (cursor) {
      array.push(cursor.value)
      cursor.continue()
    } else if (cb) {
      cb(array)
    }
  }

  cursorRequest.onerror = draftback.indexedDB.onerror
  return array
}

draftback.indexedDB.clearAllRevisions = function() {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["revision"], "readwrite")
  var store = trans.objectStore("revision")
  
  var request = store.clear()
  
  request.onsuccess = function(e) {
    console.log('successfully cleared revisions')
  }

  request.onerror = function(e) {
    console.log(e);
  }
}

draftback.clearAllRevisionsByDocId = function(docId, callback) {
  draftback.indexedDB.lastRevisionByDocId(docId, function(rev) {
    draftback.indexedDB.clearAllRevisionsByDocId(docId, rev.seq, function() {
      draftback.indexedDB.lastSessionByDocId(docId, function(sess) {
        draftback.indexedDB.clearAllSessionsByDocId(docId, sess.seq, function() {
          draftback.indexedDB.clearTimelineByDocId(docId, callback)
        })
      })
    })
  });
}

draftback.getStatsData = function(docId, callback) {
  var data = {}
  
  draftback.indexedDB.getAllSessionsByDocId(docId, function(sessions) {
    data['sessions'] = sessions
    draftback.indexedDB.getTimeline(docId, function(timeline) {
      data['timeline'] = timeline
      
      callback(data)
    })
  });
}

draftback.indexedDB.clearAllRevisionsByDocId = function(docId, upper, cb) {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["revision"], "readwrite")
  var store = trans.objectStore("revision")
  var index = store.index("docseq")
  
  var lowerBound = [docId, 0]
  var upperBound = [docId, upper]
  var range = IDBKeyRange.bound(lowerBound, upperBound)
  
  var pdestroy = index.openKeyCursor(range)
  pdestroy.onsuccess = function() {
    var cursor = pdestroy.result
    if (cursor) {
      store.delete(cursor.primaryKey)
      cursor.continue()
    } else if (cb) {
      cb()
    }
  }  
}

draftback.indexedDB.clearAllSessionsByDocId = function(docId, upper, cb) {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["session"], "readwrite")
  var store = trans.objectStore("session")
  var index = store.index("sessdocseq")
  
  var lowerBound = [docId, 0]
  var upperBound = [docId, upper]
  var range = IDBKeyRange.bound(lowerBound, upperBound)
  
  var pdestroy = index.openKeyCursor(range)
  pdestroy.onsuccess = function() {
    var cursor = pdestroy.result
    if (cursor) {
      store.delete(cursor.primaryKey)
      cursor.continue()
    } else if (cb) {
      cb()
    }
  }
}

draftback.indexedDB.getTimeline = function(docId, cb) {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["timeline"], "readwrite")
  var store = trans.objectStore("timeline")

  var request = store.get(docId)

  request.onsuccess = function(event) {
    cb(request.result)
  }

  request.onerror = draftback.indexedDB.onerror
  return null
}

draftback.indexedDB.clearTimelineByDocId = function(docId, cb) {
  var db = draftback.indexedDB.db
  var trans = db.transaction(["timeline"], "readwrite")
  var store = trans.objectStore("timeline")
    
  var request = store.delete(docId)
  request.onsuccess = function(event) {
    cb()
  }
  
  request.onerror = draftback.indexedDB.onerror
  return null
}

function initDB() {
  draftback.indexedDB.open()
}

chrome.runtime.onMessage.addListener(
  function(request, sender, sendResponse) {
    switch(request.msg) {
      case 'changelog':
        draftback.buildRevisions(request.docId, request.changelog)

        sendResponse('ok')
        break;
      
      case 'get-revisions':
        draftback.indexedDB.getRevisionsByDocIdWithOffsetAndLimit(request.docId, request.offset, request.limit, function(revs) {
          sendResponse($.map(revs, function(r, i) { return {timestamp: r.timestamp, content: r.rendered, revno: r.revno} }))
        });
        return true;
        
        break;
        
      case 'get-last-revision':
        draftback.indexedDB.lastRevisionByDocId(request.docId, function(rev) {
          sendResponse(rev)
        });
        return true;
        
        break;
        
      case 'kill':
        draftback.kills[request.docId] = true
        break;
      
      case 'clear-killed':
        delete draftback.kills[request.docId]
        break;
      
      case 'delete':
        draftback.clearAllRevisionsByDocId(request.docId, function() {
          chrome.tabs.query({url: '*://docs.google.com/*/' + request.docId + '/edit'}, function(tabs) {
            chrome.tabs.sendMessage(tabs[0].id, {msg: 'cleared'}, function(response) {});
          });
        })
        break;
      
      case 'preview':
        chrome.tabs.create({url: "/playback.html?docId=" + request.docId})
        break;
        
      case 'set-author-map':
        draftback.authorMaps[request.docId] = request.authorMap;
        break;
        
      case 'stats':
        chrome.tabs.create({url: "/stats.html?docId=" + request.docId})
        break;
        
      case 'get-stats-data':
        // Beginning?: recalling the frustration of repeated debugging attempts trying to get this to return something on the stats page?
        draftback.getStatsData(request.docId, function(data) {
          sendResponse({sessions: data.sessions, authorMap: draftback.authorMaps[request.docId], timeline: data.timeline})
        })
        return true;
      
        break;

      default:
        console.log(request)
    }
});

window.addEventListener("DOMContentLoaded", initDB, false)

// Playback memes:
// - Suppose each letter were tagged with the revision that introduced it. Then could we highlight a phrase and ask where it came from? And watch its playback? What would that even mean, for paragraphs that were composites?

// What if all *contiguous* edits were compiled into little playback streams, and clicking any one of those contiguous revisions would take you to that stream, which would be like its origin story?

// Every letter has "the set of all revisions which touched me"?

// Export to CSV.

// Migrate to an event page: https://developer.chrome.com/extensions/event_pages.